@@ -5,7 +5,7 @@
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
 ;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
 ;; URL: https://github.com/protesilaos/denote-sequence
-;; Version: 0.1.1
+;; Version: 0.2.0
 ;; Package-Requires: ((emacs "28.1") (denote "4.0.0"))
 
 ;; This file is NOT part of GNU Emacs.
@@ -53,9 +53,6 @@
 
 ;;; Code:
 
-;; TODO 2025-01-08: Test whether the built-in hierarchy.el can be used
-;; to present the sequences in a nice way.  What do we need and how
-;; exactly do we use that library.
 (require 'denote)
 
 (defgroup denote-sequence ()
@@ -384,7 +381,6 @@ Do not actually try to create a new child, as that is the duty of
 depth given SEQUENCE."
   (pcase-let* ((`(,sequence . ,scheme) (denote-sequence-and-scheme-p sequence))
                (components (denote-sequence-split sequence))
-               (butlast (butlast components))
                (last-component (car (nreverse components)))
                (new-depth (cond
                            ((eq scheme 'numeric) "1")
@@ -499,7 +495,7 @@ With optional FILES, operate on them, else use the return value of
                        (components (denote-sequence-split sequence))
                        ((>= depth (length components))))
              file))
-         (denote-sequence-get-all-files files))))
+       (or files (denote-sequence-get-all-files)))))
 
 (defun denote-sequence-get-all-sequences (&optional files)
   "Return all sequences in `denote-directory-files'.
@@ -564,31 +560,33 @@ Also see `denote-sequence-sort-files'."
   (sort
    sequences
    (lambda (sequence1 sequence2)
-       (string<
-        (denote-sequence--pad sequence1 'all)
-        (denote-sequence--pad sequence2 'all)))))
+     (string<
+      (denote-sequence--pad sequence1 'all)
+      (denote-sequence--pad sequence2 'all)))))
+
+(defun denote-sequence--file-smaller-p (file1 file2)
+  "Return non-nil if FILE1 has a smaller sequence than FILE2."
+  (let ((sequence1 (denote-retrieve-filename-signature file1))
+        (sequence2 (denote-retrieve-filename-signature file2)))
+    (string<
+     (denote-sequence--pad sequence1 'all)
+     (denote-sequence--pad sequence2 'all))))
 
 (defun denote-sequence-sort-files (files-with-sequence)
   "Sort FILES-WITH-SEQUENCE according to their sequence.
 Also see `denote-sequence-sort-sequences'."
-  (sort
-   files-with-sequence
-   (lambda (file-with-sequence-1 file-with-sequence-2)
-     (let ((s1 (denote-retrieve-filename-signature file-with-sequence-1))
-           (s2 (denote-retrieve-filename-signature file-with-sequence-2)))
-       (string<
-        (denote-sequence--pad s1 'all)
-        (denote-sequence--pad s2 'all))))))
+  (sort files-with-sequence #'denote-sequence--file-smaller-p))
 
 (defun denote-sequence--get-largest-by-order (sequences type)
   "Sort SEQUENCES of TYPE to get largest in order, using `denote-sequence--pad'."
   (car
    (reverse
-    (sort sequences
-          (lambda (s1 s2)
-            (string<
-             (denote-sequence--pad s1 type)
-             (denote-sequence--pad s2 type)))))))
+    (sort
+     sequences
+     (lambda (sequence1 sequence2)
+       (string<
+        (denote-sequence--pad sequence1 type)
+        (denote-sequence--pad sequence2 type)))))))
 
 (defun denote-sequence--string-length-sans-delimiter (string)
   "Return length of STRING without the equals sign."
@@ -743,8 +741,7 @@ returned by `denote-sequence-get-all-files'."
                            (denote-sequence-get-all-files-with-prefix prefix files)))))
     (pcase type
       ('all-parents
-       (let ((parent-files nil)
-             (butlast (butlast components))
+       (let ((butlast (butlast components))
              (found-files (denote-sequence-get-all-files files))
              (likely-parents nil))
          (while (>= (length butlast) 1)
@@ -825,7 +822,7 @@ completion candidates.  Else use `denote-sequence-get-all-files'."
                     (denote--completion-table 'file relative-files)
                     nil :require-match
                     nil 'denote-sequence-file-history)))
-      (expand-file-name input (denote-directory))
+      (expand-file-name input (car (denote-directories)))
     (error "There are no sequence notes in the `denote-directory'")))
 
 ;;;###autoload
@@ -956,11 +953,12 @@ FILES-WITH-SEQUENCES are siblings of SEQUENCE."
     (error "Cannot have a file path that satisfies `denote-sequence-file-p' while using sequence `%s'" sequence)))
 
 (defun denote-sequence-find-next-prev-sibling-subr (next-or-previous sequence relatives)
-  "Subroutine for `denote-sequence-find-next-sibling' and `denote-sequence-find-previous-sibling'.
-The NEXT-OR-PREVIOUS is the direction to move towards.  It is the symbol
-`next' or `previous'.  SEQUENCE is the one to find siblings for.
-RELATIVES is a list of files that are already known to pertain to
-SEQUENCE."
+  "Find next or previous sibling.
+Do the work for `denote-sequence-find-next-sibling' and
+`denote-sequence-find-previous-sibling'.  The NEXT-OR-PREVIOUS is the
+direction to move towards.  It is the symbol `next' or `previous'.
+SEQUENCE is the one to find siblings for.  RELATIVES is a list of files
+that are already known to pertain to SEQUENCE."
   (let ((relatives (or relatives (denote-sequence-get-relative sequence 'siblings))))
     (if-let* ((_ relatives)
               (next-in-line (denote-sequence--infer-sibling sequence next-or-previous))
@@ -1090,6 +1088,20 @@ Use optional PREFIX and DEPTH to format the string accordingly."
      (t
       (format "*Denote sequences, %s*" time)))))
 
+(defun denote-sequence--get-interactive-for-prefix-and-depth ()
+  "Return interactive list of arguments for `denote-sequence-dired' and related."
+  (let ((arg (prefix-numeric-value current-prefix-arg)))
+     (cond
+      ((= arg 16)
+       (list
+        (denote-sequence-prompt "Limit to files that extend SEQUENCE (empty for all)")
+        (denote-sequence-depth-prompt)))
+      ((= arg 4)
+       (list
+        (denote-sequence-prompt "Limit to files that extend SEQUENCE (empty for all)")))
+      (t
+       nil))))
+
 ;;;###autoload
 (defun denote-sequence-dired (&optional prefix depth)
   "Produce a Dired listing of all sequence notes.
@@ -1101,30 +1113,19 @@ With optional DEPTH as a number, limit the list to files whose sequence
 is that many levels deep.  For example, 1=1=2 is three levels deep.
 
 For a more specialised case, see `denote-sequence-find-relatives-dired'."
-  (interactive
-   (let ((arg (prefix-numeric-value current-prefix-arg)))
-     (cond
-      ((= arg 16)
-       (list
-        (denote-sequence-prompt "Limit to files that extend SEQUENCE (empty for all)")
-        (denote-sequence-depth-prompt)))
-      ((= arg 4)
-       (list
-        (denote-sequence-prompt "Limit to files that extend SEQUENCE (empty for all)")))
-      (t
-       nil))))
-  (pcase-let* ((relative-p (denote-has-single-denote-directory-p))
-               (files-fn `(lambda ()
-                            (let* ((files (if ,prefix
-                                              (denote-sequence-get-all-files-with-prefix ,prefix)
-                                            (denote-sequence-get-all-files)))
-                                   (files-with-depth (if ,depth
-                                                         (denote-sequence-get-all-files-with-max-depth ,depth files)
-                                                       files))
-                                   (files-sorted (denote-sequence-sort-files files-with-depth)))
-                              (if ,relative-p
-                                  (mapcar #'file-relative-name files-sorted)
-                                files-sorted)))))
+  (interactive (denote-sequence--get-interactive-for-prefix-and-depth))
+  (let* ((relative-p (denote-has-single-denote-directory-p))
+         (files-fn (lambda ()
+                     (let* ((files (if (and prefix (not (string-blank-p prefix)))
+                                       (denote-sequence-get-all-files-with-prefix prefix)
+                                     (denote-sequence-get-all-files)))
+                            (files-with-depth (if depth
+                                                  (denote-sequence-get-all-files-with-max-depth depth files)
+                                                files))
+                            (files-sorted (denote-sequence-sort-files files-with-depth)))
+                       (if relative-p
+                           (mapcar #'file-relative-name files-sorted)
+                         files-sorted)))))
     (if-let* ((directory (if relative-p ; see comment in `denote-file-prompt'
                              (car (denote-directories))
                            (denote-directories-get-common-root)))
@@ -1148,7 +1149,7 @@ Also see `denote-sequence-dired'."
                                         all-children
                                         children))))
   (if-let* ((sequence (denote-sequence-file-p buffer-file-name)))
-      (if-let* ((default-directory (denote-directory))
+      (if-let* ((default-directory (car (denote-directories)))
                 (relatives (ensure-list (denote-sequence-get-relative sequence type)))
                 (files-sorted (denote-sequence-sort-files relatives)))
           (dired (cons (format-message "*`%s' type relatives of `%s'" type sequence)
@@ -1161,7 +1162,7 @@ Also see `denote-sequence-dired'."
 The path is that of the special Org buffer (like `org-capture'), the
 file at point in a Dired buffer, or the variable `buffer-file-name'."
   (if (denote--file-type-org-extra-p)
-      denote-last-path-after-rename
+      denote-last-path
     (denote--rename-dired-file-or-current-file-or-prompt)))
 
 ;; TODO 2025-01-14: We need to have an operation that reparents
@@ -1245,5 +1246,185 @@ CHECK THE RESULTING SEQUENCES FOR DUPLICATES."
       (denote-rename-file file 'keep-current 'keep-current new-sequence 'keep-current 'keep-current)))
   (denote-update-dired-buffers))
 
+;;;; Display a hierarchy
+
+(defgroup denote-sequence-hierarchy ()
+  "Hierarchy view of Denote sequences."
+  :group 'denote
+  :group 'denote-sequence
+  :link '(info-link "(denote) top")
+  :link '(info-link "(denote-sequence) top")
+  :link '(url-link :tag "Denote homepage" "https://protesilaos.com/emacs/denote")
+  :link '(url-link :tag "Denote Sequence homepage" "https://protesilaos.com/emacs/denote-sequence"))
+
+(defcustom denote-sequence-hierarchy-indentation 2
+  "Number of spaces to indent by depth in `denote-sequence-view-hierarchy'."
+  :type 'natnum
+  :package-version '(denote . "0.3.0")
+  :group 'denote-sequence-hierarchy)
+
+(defun denote-sequence--format-hierarchy-entry (indent sequence title keywords)
+  "Format hierarchy entry to include INDENT, SEQUENCE, TITLE, and KEYWORDS."
+  (let* ((indent (propertize indent
+                             'cursor-sensor-functions
+                             (list
+                              (lambda (&rest _)
+                                (re-search-forward "[[:alnum:]]" nil t)
+                                (forward-char -1)))))
+         (sequence (propertize sequence 'denote-sequence-hierarchy-sequence-text t))
+         (entry (format "%s%s" indent sequence))
+         (append-fn (lambda (new prefix property)
+                      (when (and new (not (string-blank-p new)))
+                        (setq entry (format "%s %s%s" entry prefix (propertize new property t)))))))
+    (funcall append-fn title "" 'denote-sequence-hierarchy-title-text)
+    (funcall append-fn keywords "_" 'denote-sequence-hierarchy-keywords-text)
+    entry))
+
+(defun denote-sequence--hierarchy-insert (file)
+  "Insert FILE in the hierarchy with indentation matching the sequence depth."
+  (condition-case data
+      (let* ((title (denote-retrieve-title-or-filename file (denote-filetype-heuristics file)))
+             (keywords (denote-retrieve-filename-keywords file))
+             (sequence (denote-retrieve-filename-signature file))
+             (depth (denote-sequence-depth sequence))
+             (indent (if (eq depth 1)
+                         ""
+                       (make-string (* (- depth 1) denote-sequence-hierarchy-indentation) ? )))
+             (inhibit-read-only t)
+             (entry (denote-sequence--format-hierarchy-entry indent sequence title keywords)))
+        (insert (propertize entry
+                            'denote-sequence-hierarchy-level depth
+                            'denote-sequence-hierarchy-file file))
+        (insert "\n"))
+    (error (message "Failed `denote-sequence--hierarchy-insert' with data: %s" data))))
+
+(defun denote-sequence-hierarchy-get-level ()
+  "Return the outline level at point."
+  (let ((position (point)))
+    (or (get-text-property position 'denote-sequence-hierarchy-level)
+        (user-error "No outline level found at position `%s'" position))))
+
+(defun denote-sequence-hierarchy-find-file (position)
+  "Find the file at POSITION in `denote-sequence-view-hierarchy' buffer.
+When called interactively POSITION is the current `point'."
+  (interactive (list (point)))
+  (if-let* ((file (get-text-property position 'denote-sequence-hierarchy-file)))
+      (funcall denote-open-link-function file)
+    (user-error "No file found at position `%s'" position)))
+
+(defun denote-sequence--hierarchy-get-buffer (prefix depth)
+  "Return buffer for `denote-sequence-view-hierarchy'.
+PREFIX and DEPTH are used to derive the name of the buffer as well as to
+set the `revert-buffer-function'."
+  (let* ((name (format-message "*denote-sequence-hierarchy with prefix `%s'; depth `%s'*" (or prefix "ALL") (or depth "ALL")))
+         (buffer (get-buffer-create name))
+         (inhibit-read-only t))
+    (with-current-buffer buffer
+      (erase-buffer)
+      (setq-local revert-buffer-function
+                  (lambda (_ignore-auto _no-confirm)
+                    (denote-sequence-view-hierarchy prefix depth))))
+    buffer))
+
+(declare-function outline-cycle "outline" (&optional event))
+(declare-function outline-cycle-buffer "outline" (&optional level))
+(declare-function outline-forward-same-level "outline" (arg))
+(declare-function outline-backward-same-level "outline" (arg))
+(declare-function outline-next-visible-heading "outline" (arg))
+(declare-function outline-previous-visible-heading "outline" (arg))
+
+;; TODO 2025-11-19: Review which keybindings we need to cover the
+;; basic use-case.  I do not want to have a million options here.
+(defvar denote-sequence-hierarchy-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "RET") #'denote-sequence-hierarchy-find-file)
+    (define-key map (kbd "TAB") #'outline-cycle)
+    (define-key map (kbd "S-TAB") #'outline-cycle-buffer)
+    (define-key map (kbd "<backtab>") #'outline-cycle-buffer)
+    (define-key map (kbd "g") #'revert-buffer)
+    (define-key map (kbd "f") #'outline-forward-same-level)
+    (define-key map (kbd "b") #'outline-backward-same-level)
+    (define-key map (kbd "n") #'outline-next-visible-heading)
+    (define-key map (kbd "p") #'outline-previous-visible-heading)
+    map)
+  "Key map for `denote-sequence-hierarchy-mode'.")
+
+(defun denote-sequence--hierarchy-face-matcher-subr (property)
+  "Search forward for PROPERTY and return match data."
+  (when-let* ((properties (text-property-search-forward property))
+              (beginning (prop-match-beginning properties))
+              (end (prop-match-end properties)) )
+    (set-match-data (list beginning end))
+    (point)))
+
+;; FIXME 2025-12-01: `text-property-search-forward' does not have a
+;; concept of LIMIT like `re-search-forward'.  Maybe this approach of
+;; putting text properties as anchors and then searching for them is
+;; not a good idea.
+(defun denote-sequence--hierarchy-face-matcher-sequence (_limit)
+  "Font lock matcher for sequences using LIMIT."
+  (denote-sequence--hierarchy-face-matcher-subr 'denote-sequence-hierarchy-sequence-text))
+
+(defun denote-sequence--hierarchy-face-matcher-title (_limit)
+  "Font lock matcher for titles using LIMIT."
+  (denote-sequence--hierarchy-face-matcher-subr 'denote-sequence-hierarchy-title-text))
+
+(defun denote-sequence--hierarchy-face-matcher-keywords (_limit)
+  "Font lock matcher for keywords using LIMIT."
+  (denote-sequence--hierarchy-face-matcher-subr 'denote-sequence-hierarchy-keywords-text))
+
+(defvar denote-sequence-hierarchy-font-lock-keywords
+  '((denote-sequence--hierarchy-face-matcher-sequence
+     (0 'denote-faces-signature))
+    (denote-sequence--hierarchy-face-matcher-title
+     (0 'denote-faces-title))
+    (denote-sequence--hierarchy-face-matcher-keywords
+     (0 'denote-faces-keywords)))
+  "Font lock keywords for `denote-sequence-hierarchy-mode'.")
+
+(define-derived-mode denote-sequence-hierarchy-mode text-mode "Denote Hierarchy"
+  "Major mode for `denote-sequence-view-hierarchy' buffers."
+  :interactive nil
+  (setq-local font-lock-defaults '(denote-sequence-hierarchy-font-lock-keywords))
+  (setq-local outline-regexp "[\s[:alnum:]]+")
+  (setq-local outline-level #'denote-sequence-hierarchy-get-level)
+  (setq-local outline-minor-mode-highlight 'append)
+  (setq-local outline-minor-mode-cycle t)
+  (setq-local outline-minor-mode-use-buttons nil)
+  (setq-local buffer-read-only t)
+  (cursor-sensor-mode 1)
+  (outline-minor-mode 1))
+
+;;;###autoload
+(defun denote-sequence-view-hierarchy (&optional prefix depth)
+  "Show a hierachy of sequences.
+With optional PREFIX string, show only files whose sequence matches it.
+When called interactively, prompt for PREFIX, which is a file whose
+sequence is used.
+
+With optional DEPTH as a number, limit the list to files whose sequence
+is that many levels deep.  For example, 1=1=2 is three levels deep.
+When called interactively, prompt for the depth.
+
+In interactive use, PREFIX is the single universal argument, while DEPTH
+is the double universal argument.  In this case, PREFIX can be an empty
+string, which means to not use a prefix as a restriction."
+  (interactive (denote-sequence--get-interactive-for-prefix-and-depth))
+  (if-let* ((files-with-prefix (if (and prefix (not (string-blank-p prefix)))
+                                   (denote-sequence-get-all-files-with-prefix prefix)
+                                 (denote-sequence-get-all-files)))
+            (files (if depth
+                       (denote-sequence-get-all-files-with-max-depth depth files-with-prefix)
+                     files-with-prefix)))
+      (let* ((buffer (denote-sequence--hierarchy-get-buffer prefix depth))
+             (sorted (denote-sequence-sort-files files)))
+        (with-current-buffer buffer
+          (dolist (file sorted)
+            (denote-sequence--hierarchy-insert file))
+          (goto-char (point-min))
+          (denote-sequence-hierarchy-mode))
+        (display-buffer buffer))
+    (user-error "No sequences found")))
+
 (provide 'denote-sequence)
 ;;; denote-sequence.el ends here
